;		Forth Interest Group  8086 FORTH 
;
;	   Adapted to run under Microsoft's MS-DOS 8086 operating
;	system by:
;
;		J. E. Smith
;		Univ. of Pennsylvania, Dept. of Chemistry
;		250 S. 33rd St.
;		Philadelphia, PA 19104 .
;
;	   Additional modifications and enhancements
;	as described below were also implemented by Mr. Smith.
;	These changes are more fully described in a text file
;	FORTH.DOC which should accompany this source code.
;
;	   This listing is placed in the public domain, and may
;	be freely distributed.
;
;
;	Current Source Version:
;
;	1.01	06-02-82	First to assemble with no errors;
;				all CPM/86 code, but 86-DOS ASM
;				source format.
;	1.02	06-02-82	Deleted all CPM/86 dependant code,
;					substituted 86-DOS calls
;					for console i/o.
;				Changed R/W to RAM simulation.
;	1.03	06-11-82	First working version !  Some minor
;					aesthetic modifications.
;	1.10	06-12-82	Initial disk-based version.
;	1.1B	06-22-82	Configured to use 64K and 2 screens.
;					Set ^C to cause warm start.
;	1.2A	07-02-82	Modified to word align pointers.
;					Aside from assembler source
;					alignment, the following FORTH
;					words were modified:
;					(FIND),PFA,NFA,and CREATE.
;	1.2B	07-08-82	1+, 2+ changed to CODE; added 1-, 2-.
;	1.2C	07-14-82	Added (ARRAY), (2ARR), and (XOF)
;	1.2D	07-18-82	Added (CARR), (2CARR) and PRINTER for
;					echo to list output.
;	1.2E	08-18-82	Added :@, :!, :C@, :C!, MYSEG,
;					DATE@, DATE!, TIME@, TIME!.
;				Changed ^C to use (ABORT).
;				Replaced all parameters with symbols
;					defined by EQU at the start.
;---------------------------------------------------------------------
;	1.2E distributed as version 1.0
;---------------------------------------------------------------------
;	1.2F	15.10.2020	MASMized by Zbigniew Baniewski
; - replaced defaults for 8" floppy diskette by values for 3.5"/5.25" HD disks
; (not tested yet how "Forth blocks" work on floppies using these new settings)
; - replaced a few labels with different ones (the former were "reserved words")
;   DD -> DODE  DUP -> DUPE  WIDTH -> WYDTH  FLD -> FLDD  CLD -> SYSINIT
;   ABS -> ABSO  CLD1 -> COLDCFA  LEAVE -> LLEAVE  AT -> JAT (for "just AT")
; - fixed addressing by replacing all occurences of "****-$-2" with "****-$"
; - replaced "Backspace" code in boot-up literals
; - modified POUT, PKEY and ID. code to allow use of "high ASCII" characters
;   (but don't use "high ASCII" in word names though; it doesn't work yet)
; - added a few comments
; - added:  .S  PAGE  AT-XY ( x y -- )  GET-XY ( -- x y )  MSEC ( msec -- ? )
;   (MSEC - delay word, returns 0 if delay was done)

; This file can be processed by MASM or A86 with no problem whatsoever
; (old version kept in the archive for reference and comparison)

; Version numbering and ASCII equates:


FIGREL		EQU	1
FIGREV		EQU	0
USRVER		EQU	0

ABL		EQU	20H
ACR		EQU	0DH
ADOT		EQU	2EH
BELL		EQU	07H
BSIN		EQU	7FH
BSOUT		EQU	08H
DLE		EQU	10H
LF		EQU	0AH
FF		EQU	0CH

; Memory allocation parameters:

EM		EQU	0000		;64K top of memory + 1
NSCR		EQU	2		;No. of 1024 byte screens
KBBUF		EQU	1024		;No. of bytes per block
US		EQU	40H		;User area size ( in bytes )
RTS		EQU	0A0H		;Return stack/TIB size

CO		EQU	KBBUF+4		;No. bytes per block buffer
NBUF		EQU	2		;No. of block buffers =
					; NSCR*1024 / KBBUF
BUF1		EQU	0F7F8H		;Addr. of first block buffer =
					; EM - CO*NBUF
INITR0		EQU	BUF1-US		;Start of return stack (R0)
INITS0		EQU	INITR0-RTS	;Start of param. stack (S0)

; Disk parameters:
; ============= old settings for 8" floppies ===============
;TRKS		EQU	77		;Tracks on 8" disk
;SPT2		EQU	52		;8" Double density sectors/track
;SPT1		EQU	26		;8" Single density sectors/track
;SPDRV2		EQU	3744		;8" Double density sectors/drive
;SPDRV1		EQU	1872		;8" Single density sectors/drive
;BPS		EQU	128		;Bytes/sector

;SPBL		EQU	1		;Sectors/block=KBBUF/BPS
;BPSC		EQU	8		;Blocks/screen=1024/KBBUF
;MXDRV		EQU	2		;Max. number of disk drives
;DODE		EQU	0		;Density(0=single,1=double)

; ============= new settings for 3.5" (A:) and 5.25" (B:) HD floppies ========
TRKS		EQU	80		; Tracks (per side) on HD diskettes
SPT2		EQU	18		; 3.5" high density sectors/track
SPT1		EQU	15		; 5.25" high density sectors/track
SPDRV2		EQU	2880		; 3.5" high density sectors/drive
SPDRV1		EQU	2400		; 5.25" high density sectors/drive
BPS		EQU	512		; Bytes/sector

SPBL		EQU	2		;Sectors/block=KBBUF/BPS
BPSC		EQU	1		;Blocks/screen=1024/KBBUF
MXDRV		EQU	2		;Max. number of disk drives
DODE		EQU	1		;Diskette size (0 = 5.25", 1 = 3.5")
; ("memory allocation parameters" above have been recalculated accordingly!)

codeSegm	SEGMENT 'CODE'
		ASSUME CS:codeSegm, DS:codeSegm

		ORG 100h

main		PROC NEAR

ORIG: 		NOP		; (ORIGIN)
		JMP	SYSINIT
		NOP
		JMP	WRM
; boot-up literals (initial values for system variables)
		DB	FIGREL
		DB	FIGREV
		DB	USRVER
		DB	0EH
		DW	TASK - 8; (ORIGIN + 0Ch) last definition in FORTH voc.
		DW	BSOUT	; (ORIGIN + 0Eh) "Backspace" code
		DW	INITR0

		DW	INITS0	; (ORIGIN + 12h) S0
		DW	INITR0	; R0
		DW	INITS0	; TIB
		DW	32	; WIDTH
		DW	0	; WARNING
		DW	INITDP	; FENCE
		DW	INITDP	; DP
		DW	FORTH+6	; VOC-LINK

		DW	05H,0B326H		;"8086" ( in base 36 ! )
UP 		DW	INITR0	; (ORIGIN + 26h) start-address of user area
RPP 		DW	INITR0
BIP 		DW	0
BIPE 		DW	0

TNEXT: 		PUSHF
		PUSH	AX
		MOV	AX,[BIP]
		OR	AX,AX
		JZ	TNEXT2
		CMP	AX,-1
		JZ	TNEXT1
		CMP	AX,SI
		JZ	TNEXT1
		JA	TNEXT2
		MOV	AX,[BIPE]
		OR	AX,AX
		JZ	TNEXT2
		CMP	AX,SI
		JB	TNEXT2
TNEXT1: 	POP	AX
		POPF
BREAK: 		JP	TNEXT3
TNEXT2: 	POP	AX
		POPF
TNEXT3: 	LODSW
		MOV	BX,AX
		JP	NEXT1
DPUSH: 		PUSH	DX
APUSH: 		PUSH	AX
NEXT: 		LODSW		; [DS:SI] -> AX    SI = "IP"  DX = "W"?
		MOV	BX,AX
NEXT1: 		MOV	DX,BX
		INC	DX
		JMP	[BX]	; BX = "X" reg.? BP = top of return stack?

	ALIGN 2
DP0 		DB	83H,"LI","T"+80h
		DW	0
LIT 		DW	$ + 2
		LODSW
		JMP	APUSH

	ALIGN 2
		DB	87H,"EXECUT","E"+80h
		DW	LIT - 6
EXEC 		DW	$ + 2
		POP	BX
		JMP	NEXT1

	ALIGN 2
		DB	86H,"BRANC","H"+80h
	ALIGN 2
		DW	EXEC - 10
BRAN 		DW	$ + 2
BRAN1:		ADD	SI,[SI]
		JMP	NEXT

	ALIGN 2
		DB	87H,"0BRANC","H"+80h
		DW	BRAN - 10
ZBRAN 		DW	$ + 2
		POP	AX
		OR	AX,AX
		JZ	BRAN1
		INC	SI
		INC	SI
		JMP	NEXT

	ALIGN 2
		DB	86H,"(LOOP",")"+80h
	ALIGN 2
		DW	ZBRAN - 10
XLOOP 		DW	$ + 2
		MOV	BX,1
XLOO1: 		ADD	[BP],BX
		MOV	AX,[BP]
		SUB	AX,[BP+2]
		XOR	AX,BX
		JS	BRAN1

		ADD	BP,4
		INC	SI
		INC	SI
		JMP	NEXT

	ALIGN 2
		DB	87H,"(+LOOP",")"+80h
		DW	XLOOP - 10
XPLOO 		DW	$ + 2
		POP	BX
		JMP	XLOO1

	ALIGN 2
		DB	84H,"(DO",")"+80h
	ALIGN 2
		DW	XPLOO - 10
XDO 		DW	$ + 2
		POP	DX
		POP	AX
		XCHG	BP,SP
		PUSH	AX
		PUSH	DX
		XCHG	BP,SP
		JMP	NEXT

;************************
;*			*
;*	(XOF)		*
;*			*
;************************
;
;	Code added for Dr. Eaker's CASE construct
;	After John Cassady's 8080 code in FD 3:187 1982
;	(jes ver1.2C,1982)

	ALIGN 2
		DB	85H,"(XOF",")"+80h
		DW	XDO - 8
XOF		DW	$ + 2
		POP	BX		;BX := case tag
		POP	AX		;AX := search tag
		CMP	AX,BX		;This one ?
		JE	XOF1		;Yes...
		PUSH	AX		;No, save search tag,
		JMP	BRAN1		;   and check the next case.
XOF1:		INC	SI		;...skip the branch offset,
		INC	SI		;   and
		JMP	NEXT		;   don't save the search tag.


	ALIGN 2
		DB	81H,"I"+80h
		DW	XOF - 8
IDO		DW	$ + 2
		MOV	AX,[BP]
		JMP	APUSH

	ALIGN 2
		DB	85H,"DIGI","T"+80h
		DW	IDO - 4
DIGIT 		DW	$ + 2
		POP	DX
		POP	AX
		SUB	AL,'0'
		JB	DIGI2
		CMP	AL,9
		JBE	DIGI1
		SUB	AL,7
		CMP	AL,10
		JB	DIGI2
DIGI1: 		CMP	AL,DL
		JAE	DIGI2
		SUB	DX,DX
		MOV	DL,AL
		MOV	AL,1
		JMP	DPUSH
DIGI2: 		SUB	AX,AX
		JMP	APUSH

	ALIGN 2
		DB	86H,"(FIND",")"+80h
	ALIGN 2
		DW	DIGIT - 8
PFIND 		DW	$ + 2
		MOV	AX,DS
		MOV	ES,AX
		POP	BX
		POP	CX
PFIN1: 		MOV	DI,CX
		MOV	AL,[BX]
		MOV	DL,AL
		XOR	AL,[DI]
		AND	AL,3FH
		JNZ	PFIN5
PFIN2: 		INC	BX
		INC	DI
		MOV	AL,[BX]
		XOR	AL,[DI]
		ADD	AL,AL
		JNZ	PFIN5
		JNB	PFIN2

		ADD	BX,6		;Compute PFA (could be 5 or 6)
		AND	BX,0FFFEH	;Clear LSB to align

		PUSH	BX
		MOV	AX,1
		SUB	DH,DH
		JMP	DPUSH
PFIN5: 		INC	BX
		JB	PFIN6
		MOV	AL,[BX]
		ADD	AL,AL
		JMP	PFIN5

PFIN6:		INC	BX		;This could be one too many...
		AND	BX,0FFFEH	;Clear LSB to align

		MOV	BX,[BX]
		OR	BX,BX
		JNZ	PFIN1
		MOV	AX,0
		JMP	APUSH

	ALIGN 2
		DB	87H,"ENCLOS","E"+80h
		DW	PFIND - 10
ENCL 		DW	$ + 2
		POP	AX
		POP	BX
		PUSH	BX
		MOV	AH,0
		MOV	DX,-1
		DEC	BX
ENCL1: 		INC	BX
		INC	DX
		CMP	AL,[BX]
		JZ	ENCL1
		PUSH	DX
		CMP	AH,[BX]
		JNZ	ENCL2
		MOV	AX,DX
		INC	DX
		JMP	DPUSH
ENCL2: 		INC	BX
		INC	DX
		CMP	AL,[BX]
		JZ	ENCL4
		CMP	AH,[BX]
		JNZ	ENCL2
ENCL3: 		MOV	AX,DX
		JMP	DPUSH
ENCL4: 		MOV	AX,DX
		INC	AX
		JMP	DPUSH

	ALIGN 2
		DB	84H,"EMI","T"+80h
	ALIGN 2
		DW	ENCL - 10
EMIT 		DW	DOCOL
		DW	PEMIT
		DW	ONE,OUTT
		DW	PSTOR,SEMIS

	ALIGN 2
		DB	83H,"KE","Y"+80h
		DW	EMIT - 8
KEY 		DW	$ + 2
		JMP	PKEY

	ALIGN 2
		DB	89H,"?TERMINA","L"+80h
		DW	KEY - 6
QTERM 		DW	$ + 2
		JMP	PQTER

	ALIGN 2
		DB	82H,"C","R"+80h
	ALIGN 2
		DW	QTERM - 12
CR 		DW	$ + 2
		JMP	PCR

	ALIGN 2
		DB	85H,"CMOV","E"+80h
		DW	CR - 6
CMOVE 		DW	$ + 2
		CLD
		MOV	BX,SI
		POP	CX
		POP	DI
		POP	SI
		MOV	AX,DS
		MOV	ES,AX
		REP	MOVSB
		MOV	SI,BX
		JMP	NEXT

	ALIGN 2
		DB	82H,"U","*"+80h
	ALIGN 2
		DW	CMOVE - 8
USTAR 		DW	$ + 2
		POP	AX
		POP	BX
		MUL	BX
		XCHG	AX,DX
		JMP	DPUSH

	ALIGN 2
		DB	82H,"U","/"+80h
	ALIGN 2
		DW	USTAR - 6
USLAS 		DW	$ + 2
		POP	BX
		POP	DX
		POP	AX
		CMP	DX,BX
		JNB	DZERO
		DIV	BX
		JMP	DPUSH
DZERO: 		MOV	AX,-1
		MOV	DX,AX
		JMP	DPUSH

	ALIGN 2
		DB	83H,"AN","D"+80h
		DW	USLAS - 6
ANDD 		DW	$ + 2
		POP	AX
		POP	BX
		AND	AX,BX
		JMP	APUSH

	ALIGN 2
		DB	82H,"O","R"+80h
	ALIGN 2
		DW	ANDD - 6
ORR 		DW	$ + 2
		POP	AX
		POP	BX
		OR	AX,BX
		JMP	APUSH

	ALIGN 2
		DB	83H,"XO","R"+80h
		DW	ORR - 6
XORR		DW	$ + 2
		POP	AX
		POP	BX
		XOR	AX,BX
		JMP	APUSH

	ALIGN 2
		DB	83H,"SP","@"+80h
		DW	XORR - 6
SPAT 		DW	$ + 2
		MOV	AX,SP
		JMP	APUSH

	ALIGN 2
		DB	83H,"SP","!"+80h
		DW	SPAT - 6
SPSTO 		DW	$ + 2
		MOV	BX,[UP]
		MOV	SP,[BX+6]
		JMP	NEXT

	ALIGN 2
		DB	83H,"RP","@"+80h
		DW	SPSTO - 6
RPAT 		DW	$ + 2
		MOV	AX,BP
		JMP	APUSH

	ALIGN 2
		DB	83H,"RP","!"+80h
		DW	RPAT - 6
RPSTO		DW	$ + 2
		MOV	BX,[UP]
		MOV	BP,[BX+8]
		JMP	NEXT

	ALIGN 2
		DB	82H,";","S"+80h
	ALIGN 2
		DW	RPSTO - 6
SEMIS 		DW	$ + 2
		MOV	SI,[BP]
		INC	BP
		INC	BP
		JMP	NEXT

	ALIGN 2
		DB	85H,"LEAV","E"+80h
		DW	SEMIS - 6
LLEAVE 		DW	$ + 2
		MOV	AX,[BP]
		MOV	[BP+2],AX
		JMP	NEXT

	ALIGN 2
		DB	82H,">","R"+80h
	ALIGN 2
		DW	LLEAVE - 8
TOR 		DW	$ + 2
		POP	BX
		DEC	BP
		DEC	BP
		MOV	[BP],BX
		JMP	NEXT

	ALIGN 2
		DB	82H,"R",">"+80h
	ALIGN 2
		DW	TOR - 6
FROMR 		DW	$ + 2
		MOV	AX,[BP]
		INC	BP
		INC	BP
		JMP	APUSH

	ALIGN 2
		DB	81H,"R"+80h
		DW	FROMR - 6
RR 		DW	IDO + 2

	ALIGN 2
		DB	82H,"0","="+80h
	ALIGN 2
		DW	RR - 4
ZEQU 		DW	$ + 2
		POP	AX
		OR	AX,AX
		MOV	AX,1
		JZ	ZEQU1
		DEC	AX
ZEQU1:		JMP	APUSH

	ALIGN 2
		DB	82H,"0","<"+80h
	ALIGN 2
		DW	ZEQU - 6
ZLESS 		DW	$ + 2
		POP	AX
		OR	AX,AX
		MOV	AX,1
		JS	ZLESS1
		DEC	AX
ZLESS1:		JMP	APUSH

	ALIGN 2
		DB	81H,"+"+80h
		DW	ZLESS - 6
PLUS 		DW	$ + 2
		POP	AX
		POP	BX
		ADD	AX,BX
		JMP	APUSH

	ALIGN 2
		DB	82H,"D","+"+80h
	ALIGN 2
		DW	PLUS - 4
DPLUS 		DW	$ + 2
		POP	AX
		POP	DX
		POP	BX
		POP	CX
		ADD	DX,CX
		ADC	AX,BX
		JMP	DPUSH

	ALIGN 2
		DB	85H,"MINU","S"+80h
		DW	DPLUS - 6
MINUS		DW	$ + 2
		POP	AX
		NEG	AX
		JMP	APUSH

	ALIGN 2
		DB	86H,"DMINU","S"+80h
	ALIGN 2
		DW	MINUS - 8
DMINU 		DW	$ + 2
		POP	BX
		POP	CX
		SUB	AX,AX
		MOV	DX,AX
		SUB	DX,CX
		SBB	AX,BX
		JMP	DPUSH

	ALIGN 2
		DB	84H,"OVE","R"+80h
	ALIGN 2
		DW	DMINU - 10
OVER		DW	$ + 2
		POP	DX
		POP	AX
		PUSH	AX
		JMP	DPUSH

	ALIGN 2
		DB	84H,"DRO","P"+80h
	ALIGN 2
		DW	OVER - 8
DROP 		DW	$ + 2
		POP	AX
		JMP	NEXT

	ALIGN 2
		DB	84H,"SWA","P"+80h
	ALIGN 2
		DW	DROP - 8
SWAP		DW	$ + 2
		POP	DX
		POP	AX
		JMP	DPUSH

	ALIGN 2
		DB	83H,"DU","P"+80h
		DW	SWAP - 8
DUPE		DW	$ + 2
		POP	AX
		PUSH	AX
		JMP	APUSH

	ALIGN 2
		DB	84H,"2DU","P"+80h
	ALIGN 2
		DW	DUPE - 6
TDUP		DW	$ + 2
		POP	AX
		POP	DX
		PUSH	DX
		PUSH	AX
		JMP	DPUSH

	ALIGN 2
		DB	82H,"+","!"+80h
	ALIGN 2
		DW	TDUP - 8
PSTOR		DW	$ + 2
		POP	BX
		POP	AX
		ADD	[BX],AX
		JMP	NEXT

	ALIGN 2
		DB	86H,"TOGGL","E"+80h
	ALIGN 2
		DW	PSTOR - 6
TOGGL		DW	$ + 2
		POP	AX
		POP	BX
		XOR	[BX],AL
		JMP	NEXT

	ALIGN 2
		DB	81H,"@"+80h
		DW	TOGGL - 10
JAT		DW	$ + 2
		POP	BX
		MOV	AX,[BX]
		JMP	APUSH

	ALIGN 2
		DB	82H,"C","@"+80h
	ALIGN 2
		DW	JAT - 4
CAT		DW	$ + 2
		POP	BX
		MOV	AL,[BX]
		SUB	AH,AH
		JMP	APUSH

	ALIGN 2
		DB	82H,"2","@"+80h
	ALIGN 2
		DW	CAT - 6
TAT		DW	$ + 2
		POP	BX
		MOV	AX,[BX]
		MOV	DX,[BX+2]
		JMP	DPUSH

	ALIGN 2
		DB	81H,"!"+80h
		DW	TAT - 6
STORE		DW	$ + 2
		POP	BX
		POP	AX
		MOV	[BX],AX
		JMP	NEXT

	ALIGN 2
		DB	82H,"C","!"+80h
	ALIGN 2
		DW	STORE - 4
CSTOR		DW	$ + 2
		POP	BX
		POP	AX
		MOV	[BX],AL
		JMP	NEXT

	ALIGN 2
		DB	82H,"2","!"+80h
	ALIGN 2
		DW	CSTOR - 6
TSTOR		DW	$ + 2
		POP	BX
		POP	AX
		MOV	[BX],AX
		POP	AX
		MOV	[BX+2],AX
		JMP	NEXT

;********************************************************
;*							*
;*	long fetch/store operators:	:@, :!		*
;*					:C@, :C!	*
;*					MYSEG		*
;*							*
;********************************************************

	ALIGN 2
		DB	82H,":","@"+80h
	ALIGN 2
		DW	TSTOR - 6
FARAT		DW	$ + 2
		POP	BX		;Offset
		MOV	DX,DS		;Save current segment
		POP	DS		;Segment
		MOV	AX,[BX]		;Fetch word at DS:BX
		MOV	DS,DX		;Restore segment register
		JMP	APUSH		;Return

	ALIGN 2
		DB	82H,":","!"+80h
	ALIGN 2
		DW	FARAT - 6
FARST		DW	$ + 2
		MOV	DX,DS
		POP	BX		;Offset
		POP	DS		;Segment
		POP	AX		;Data
		MOV	[BX],AX
		MOV	DS,DX
		JMP	NEXT

	ALIGN 2
		DB	83H,":C","@"+80h
		DW	FARST - 6
FARCAT		DW	$ + 2
		MOV	DX,DS
		POP	BX
		POP	DS
		MOV	AL,[BX]		; here was MOV B,AL,[BX] (?)
		XOR	AH,AH
		MOV	DS,DX
		JMP	APUSH

	ALIGN 2
		DB	83H,":C","!"+80h
		DW	FARCAT - 6
FARCST		DW	$ + 2
		MOV	DX,DS
		POP	BX
		POP	DS
		POP	AX
		MOV	[BX],AL		; here was MOV B,[BX],AL (?)
		MOV	DS,DX
		JMP	NEXT

	ALIGN 2
		DB	85H,"MYSE","G"+80h
		DW	FARCST - 6
MYSEG		DW	$ + 2
		MOV	AX,DS
		JMP	APUSH

	ALIGN 2
		DB	0C1H,":"+80h
		DW	MYSEG - 8
COLON		DW	DOCOL
		DW	QEXEC,	SCSP
		DW	CURR,	JAT
		DW	CONT,	STORE
		DW	CREAT,	RBRAC
		DW	PSCOD
DOCOL:		INC	DX
		DEC	BP
		DEC	BP
		MOV	[BP],SI
		MOV	SI,DX
		JMP	NEXT

	ALIGN 2
		DB	0C1H,";"+80h
		DW	COLON - 4
SEMI		DW	DOCOL
		DW	QCSP,	COMP
		DW	SEMIS,	SMUDG
		DW	LBRAC,	SEMIS

	ALIGN 2
		DB	84H,"NOO","P"+80h
	ALIGN 2
		DW	SEMI - 4
NOOP		DW	DOCOL,	SEMIS	; former solution

	ALIGN 2
		DB	88H,"CONSTAN","T"+80h
	ALIGN 2
		DW	NOOP - 8
CON		DW	DOCOL
		DW	CREAT,	SMUDG
		DW	COMMA,	PSCOD
DOCON:		INC	DX
		MOV	BX,DX
		MOV	AX,[BX]
		JMP	APUSH

	ALIGN 2
		DB	88H,"VARIABL","E"+80h
	ALIGN 2
		DW	CON - 12
VAR		DW	DOCOL
		DW	CON,	PSCOD
DOVAR:		INC	DX
		PUSH	DX
		JMP	NEXT

	ALIGN 2
		DB	84H,"USE","R"+80h
	ALIGN 2
		DW	VAR - 12
USER		DW	DOCOL
		DW	CON,	PSCOD
DOUSE:		INC	DX
		MOV	BX,DX
		MOV	BL,[BX]
		SUB	BH,BH
		MOV	DI,[UP]
		LEA	AX,[BX+DI]
		JMP	APUSH

;************************
;*			*
;*	(ARRAY)		*
;*			*
;************************

;	Code added to support array references.
;	Used by ARRAY to calculate the address of the
;	nth element of the array.
;	(jes ver1.2c,1982)

	ALIGN 2
		DB	87H,"(ARRAY",")"+80h
		DW	USER - 8
PARR		DW	$ + 2
		POP	BX		;BX -> SIZE
		POP	AX		;AX := n
		ADD	AX,AX		;AX := AX*2
		ADD	AX,BX		;AX -> ARRAY[n]
		ADD	AX,2		;Offset to ARRAY[0]
		JMP	APUSH

	ALIGN 2
		DB	86H,"(2ARR",")"+80h
	ALIGN 2
		DW	PARR - 10
P2ARR		DW	$ + 2
		POP	BX		;BX -> rowsize
		POP	CX		;CX := column
		POP	AX		;AX := row
		MUL	WORD PTR [BX]	;AX := row*row dim.
		ADD	AX,CX		;AX := AX + col
		ADD	AX,AX		;2 bytes per element
		ADD	AX,BX		;AX := AX+PFA
		ADD	AX,4		;Offset to ARRAY[0]
		JMP	APUSH

	ALIGN 2
		DB	86H,"(CARR",")"+80h
	ALIGN 2
		DW	P2ARR - 10
PCARR		DW	$ + 2
		POP	BX
		POP	AX
		ADD	AX,BX
		ADD	AX,2
		JMP	APUSH

	ALIGN 2
		DB	87H,"(2CARR",")"+80h
		DW	PCARR - 10
P2CAR		DW	$ + 2
		POP	BX
		POP	CX
		POP	AX
		MUL	WORD PTR [BX]
		ADD	AX,CX
		ADD	AX,BX
		ADD	AX,4
		JMP	APUSH

	ALIGN 2
		DB	81H,"0"+80h
		DW	P2CAR - 10
ZERO		DW	DOCON
		DW	0

		DB	81H,"1"+80h
		DW	ZERO - 4
ONE		DW	DOCON
		DW	1

		DB	81H,"2"+80h
		DW	ONE - 4
TWO		DW	DOCON
		DW	2

		DB	81H,"3"+80h
		DW	TWO - 4
THREE		DW	DOCON
		DW	3

		DB	82H,"B","L"+80h
	ALIGN 2
		DW	THREE - 4
BLS		DW	DOCON
		DW	20H

		DB	83H,"C/","L"+80h
		DW	BLS - 6
CSLL		DW	DOCON
		DW	64

		DB	85H,"FIRS","T"+80h
		DW	CSLL - 6
FIRST		DW	DOCON
		DW	BUF1

		DB	85H,"LIMI","T"+80h
		DW	FIRST - 8
LIMIT		DW	DOCON
		DW	EM

		DB	85H,"B/BU","F"+80h
		DW	LIMIT - 8
BBUF		DW	DOCON
		DW	KBBUF

		DB	85H,"B/SC","R"+80h
		DW	BBUF - 8
BSCR		DW	DOCON
		DW	BPSC		; 400H/KBBUF

		DB	87H,"+ORIGI","N"+80h
		DW	BSCR - 8
PORIG		DW	DOCOL
		DW	LIT,	ORIG
		DW	PLUS,	SEMIS

		DB	82H,"S","0"+80h
	ALIGN 2
		DW	PORIG - 10
SZERO		DW	DOUSE
		DW	6

		DB	82H,"R","0"+80h
	ALIGN 2
		DW	SZERO - 6
RZERO		DW	DOUSE
		DW	8

		DB	83H,"TI","B"+80h
		DW	RZERO - 6
TIB 		DW	DOUSE
		DW	10

		DB	85H,"WIDT","H"+80h
		DW	TIB - 6
WYDTH		DW	DOUSE
		DW	12

		DB	87H,"WARNIN","G"+80h
		DW	WYDTH - 8
WARN		DW	DOUSE
		DW	14

		DB	85H,"FENC","E"+80h
		DW	WARN - 10
FENCE 		DW	DOUSE
		DW	16

		DB	82H,"D","P"+80h
	ALIGN 2
		DW	FENCE - 8
DP		DW	DOUSE
		DW	18

		DB	88H,"VOC-LIN","K"+80h
	ALIGN 2
		DW	DP - 6
VOCL		DW	DOUSE
		DW	20

		DB	83H,"BL","K"+80h
		DW	VOCL - 12
BLK		DW	DOUSE
		DW	22

		DB	82H,"I","N"+80h
	ALIGN 2
		DW	BLK - 6
INN		DW	DOUSE
		DW	24

		DB	83H,"OU","T"+80h
		DW	INN - 6
OUTT		DW	DOUSE
		DW	26

		DB	83H,"SC","R"+80h
		DW	OUTT - 6
SCR		DW	DOUSE
		DW	28

		DB	86H,"OFFSE","T"+80h
	ALIGN 2
		DW	SCR - 6
OFSET		DW	DOUSE
		DW	30

		DB	87H,"CONTEX","T"+80h
		DW	OFSET - 10
CONT		DW	DOUSE
		DW	32

		DB	87H,"CURREN","T"+80h
		DW	CONT - 10
CURR		DW	DOUSE
		DW	34

		DB	85H,"STAT","E"+80h
		DW	CURR - 10
STATE		DW	DOUSE
		DW	36

		DB	84H,"BAS","E"+80h
	ALIGN 2
		DW	STATE - 8
BASE		DW	DOUSE
		DW	38

		DB	83H,"DP","L"+80h
		DW	BASE - 8
DPL		DW	DOUSE
		DW	40

		DB	83H,"FL","D"+80h
		DW	DPL - 6
FLDD		DW	DOUSE
		DW	42

		DB	83H,"CS","P"+80h
		DW	FLDD - 6
CSPP		DW	DOUSE
		DW	44

		DB	82H,"R","#"+80h
	ALIGN 2
		DW	CSPP - 6
RNUM		DW	DOUSE
		DW	46

		DB	83H,"HL","D"+80h
		DW	RNUM - 6
HLD		DW	DOUSE
		DW	48

		DB	82H,"1","+"+80h
	ALIGN 2
		DW	HLD - 6
ONEP		DW	$ + 2
		POP	AX
		INC	AX
		JMP	APUSH

	ALIGN 2
		DB	82H,"2","+"+80h
	ALIGN 2
		DW	ONEP - 6
TWOP		DW	$ + 2
		POP	AX
		INC	AX
		INC	AX
		JMP	APUSH

	ALIGN 2
		DB	82H,"1","-"+80h
	ALIGN 2
		DW	TWOP - 6
ONEM		DW	$ + 2
		POP	AX
		DEC	AX
		JMP	APUSH
	ALIGN 2
		DB	82H,"2","-"+80h
	ALIGN 2
		DW	ONEM - 6
TWOM		DW	$ + 2
		POP	AX
		DEC	AX
		DEC	AX
		JMP	APUSH
	ALIGN 2
		DB	84H,"HER","E"+80h
	ALIGN 2
		DW	TWOM - 6
HERE		DW	DOCOL
		DW	DP,	JAT,	SEMIS

		DB	85H,"ALLO","T"+80h
		DW	HERE - 8
ALLOT		DW	DOCOL
		DW	DP,	PSTOR,	SEMIS

		DB	81H,","+80h
		DW	ALLOT - 8
COMMA		DW	DOCOL
		DW	HERE,	STORE
		DW	TWO,	ALLOT,	SEMIS

		DB	82H,"C",","+80h
	ALIGN 2
		DW	COMMA - 4
CCOMM		DW	DOCOL
		DW	HERE,	CSTOR
		DW	ONE,	ALLOT,	SEMIS

		DB	81H,"-"+80h
		DW	CCOMM - 6
SUBB		DW	$ + 2
		POP	DX
		POP	AX
		SUB	AX,DX
		JMP	APUSH

	ALIGN 2
		DB	81H,"="+80h
		DW	SUBB - 4
EQUAL		DW	DOCOL
		DW	SUBB,	ZEQU,	SEMIS

		DB	81H,"<"+80h
		DW	EQUAL - 4
LESS		DW	$ + 2
		POP	DX
		POP	AX
		MOV	BX,DX
		XOR	BX,AX
		JS	LES1
		SUB	AX,DX
LES1: 		OR	AX,AX
		MOV	AX,0
		JNS	LES2
		INC	AX
LES2: 		JMP	APUSH

	ALIGN 2
		DB	82H,"U","<"+80h
	ALIGN 2
		DW	LESS - 4
ULESS		DW	DOCOL
		DW	TDUP,	XORR,	ZLESS
		DW	ZBRAN,	ULES1-$
		DW	DROP,	ZLESS,	ZEQU
		DW	BRAN,	ULES2-$
ULES1		DW	SUBB,	ZLESS
ULES2		DW	SEMIS

		DB	81H,">"+80h
		DW	ULESS - 6
GREAT		DW	DOCOL
		DW	SWAP,	LESS,	SEMIS

		DB	83H,"RO","T"+80h
		DW	GREAT - 4
ROT		DW	$ + 2
		POP	DX
		POP	BX
		POP	AX
		PUSH	BX
		JMP	DPUSH

	ALIGN 2
		DB	85H,"SPAC","E"+80h
		DW	ROT - 6
SPACE		DW	DOCOL
		DW	BLS,	EMIT,	SEMIS

		DB	84H,"-DU","P"+80h
	ALIGN 2
		DW	SPACE - 8
DDUP		DW	DOCOL
		DW	DUPE
		DW	ZBRAN,	DDUP1-$
		DW	DUPE
DDUP1		DW	SEMIS

		DB	88H,"TRAVERS","E"+80h
	ALIGN 2
		DW	DDUP - 8
TRAV		DW	DOCOL
		DW	SWAP
TRAV1		DW	OVER,	PLUS
		DW	LIT,	7FH
		DW	OVER,	CAT,	LESS
		DW	ZBRAN,	TRAV1-$
		DW	SWAP,	DROP,	SEMIS

		DB	86H,"LATES","T"+80h
	ALIGN 2
		DW	TRAV - 12
LATES		DW	DOCOL
		DW	CURR,	JAT,	JAT,	SEMIS

		DB	83H,"LF","A"+80h
		DW	LATES - 10
LFA		DW	DOCOL
		DW	LIT,	4
		DW	SUBB,	SEMIS

		DB	83H,"CF","A"+80h
		DW	LFA - 6
CFA		DW	DOCOL
		DW	TWO,	SUBB,	SEMIS

		DB	83H,"NF","A"+80h
		DW	CFA - 6
NFA		DW	DOCOL
		DW	LIT,	5		;Could be 5 or 6
		DW	SUBB
		DW	DUPE,	CAT
		DW	LIT,	80H,	ANDD,	ZEQU
		DW	ZBRAN,	NFA1-$	;MSB set, OK
		DW	ONEM			;MSB not set, adjust
NFA1		DW	LIT,	-1
		DW	TRAV,	SEMIS

		DB	83H,"PF","A"+80h
		DW	NFA - 6
PFA		DW	$ + 2
		POP	BX		;BX:=NFA
		MOV	AL,[BX]		;AL:=count
		AND	AL,1FH		;Only lowest 5 bits
		ADD	AL,6
		SUB	AH,AH
		ADD	BX,AX		;BX:=NFA+count+6
		AND	BX,0FFFEH	;Clear LSB to align
		MOV	AX,BX
		JMP	APUSH		;Save PFA

	ALIGN 2
		DB	84H,"!CS","P"+80h
	ALIGN 2
		DW	PFA - 6
SCSP		DW	DOCOL
		DW	SPAT,	CSPP
		DW	STORE,	SEMIS

		DB	86H,"?ERRO","R"+80h
	ALIGN 2
		DW	SCSP - 8
QERR		DW	DOCOL
		DW	SWAP
		DW	ZBRAN,	QERR1-$
		DW	ERROR
		DW	BRAN,	QERR2-$
QERR1		DW	DROP
QERR2		DW	SEMIS

		DB	85H,"?COM","P"+80h
		DW	QERR - 10
QCOMP		DW	DOCOL
		DW	STATE,	JAT
		DW	ZEQU,	LIT,	17
		DW	QERR,	SEMIS

		DB	85H,"?EXE","C"+80h
		DW	QCOMP - 8
QEXEC		DW	DOCOL
		DW	STATE,	JAT
		DW	LIT,	18
		DW	QERR,	SEMIS

		DB	86H,"?PAIR","S"+80h
	ALIGN 2
		DW	QEXEC - 8
QPAIR		DW	DOCOL
		DW	SUBB
		DW	LIT,	19
		DW	QERR,	SEMIS

		DB	84H,"?CS","P"+80h
	ALIGN 2
		DW	QPAIR - 10
QCSP		DW	DOCOL
		DW	SPAT,	CSPP,	JAT,	SUBB
		DW	LIT,	20
		DW	QERR,	SEMIS

		DB	88H,"?LOADIN","G"+80h
	ALIGN 2
		DW	QCSP - 8
QLOAD		DW	DOCOL
		DW	BLK,	JAT,	ZEQU
		DW	LIT,	22
		DW	QERR,	SEMIS

		DB	87H,"COMPIL","E"+80h
		DW	QLOAD - 12
COMP		DW	DOCOL
		DW	QCOMP
		DW	FROMR,	DUPE,	TWOP,	TOR
		DW	JAT,	COMMA,	SEMIS

		DB	0C1H,"["+80h
		DW	COMP - 10
LBRAC		DW	DOCOL
		DW	ZERO,	STATE,	STORE,	SEMIS

		DB	81H,"]"+80h
		DW	LBRAC - 4
RBRAC		DW	DOCOL
		DW	LIT,	0C0H
		DW	STATE,	STORE,	SEMIS

		DB	86H,"SMUDG","E"+80h
	ALIGN 2
		DW	RBRAC - 4
SMUDG		DW	DOCOL
		DW	LATES
		DW	LIT,	20H
		DW	TOGGL,	SEMIS

		DB	83H,"HE","X"+80h
		DW	SMUDG - 10
HEX		DW	DOCOL
		DW	LIT,	16
		DW	BASE,	STORE,	SEMIS

		DB	87H,"DECIMA","L"+80h
		DW	HEX - 6
DECA		DW	DOCOL
		DW	LIT,	10
		DW	BASE,	STORE,	SEMIS

		DB	87H,"(;CODE",")"+80h
		DW	DECA - 10
PSCOD		DW	DOCOL
		DW	FROMR,	LATES,	PFA
		DW	CFA,	STORE,	SEMIS

		DB	0C5H,";COD","E"+80h
		DW	PSCOD - 10
SEMIC		DW	DOCOL
		DW	QCSP
		DW	COMP,	PSCOD,	LBRAC
SEMI1		DW	NOOP
		DW	SEMIS

		DB	87H,"<BUILD","S"+80h
		DW	SEMIC - 8
BUILD		DW	DOCOL
		DW	ZERO,	CON,	SEMIS

		DB	85H,"DOES",">"+80h
		DW	BUILD - 10
DOES		DW	DOCOL
		DW	FROMR,	LATES,	PFA,	STORE
		DW	PSCOD
DODOE: 		XCHG	BP,SP
		PUSH	SI
		XCHG	BP,SP
		INC	DX
		MOV	BX,DX
		MOV	SI,[BX]
		INC	DX
		INC	DX
		PUSH	DX
		JMP	NEXT

	ALIGN 2
		DB	85H,"COUN","T"+80h
		DW	DOES - 8
COUNT		DW	DOCOL
		DW	DUPE,	ONEP,	SWAP,	CAT,	SEMIS

		DB	84H,"TYP","E"+80h
	ALIGN 2
		DW	COUNT - 8
TYPES		DW	DOCOL
		DW	DDUP
		DW	ZBRAN,	TYPE1-$
		DW	OVER,	PLUS
		DW	SWAP,	XDO
TYPE2		DW	IDO,	CAT,	EMIT
		DW	XLOOP,	TYPE2-$
		DW	BRAN,	TYPE3-$
TYPE1		DW	DROP
TYPE3		DW	SEMIS

		DB	89H,"-TRAILIN","G"+80h
		DW	TYPES - 8
DTRAI		DW	DOCOL
		DW	DUPE,	ZERO,	XDO
DTRA1		DW	OVER,	OVER,	PLUS
		DW	ONE,	SUBB,	CAT
		DW	BLS,	SUBB
		DW	ZBRAN,	DTRA2-$
		DW	LLEAVE
		DW	BRAN,	DTRA3-$
DTRA2		DW	ONE,	SUBB
DTRA3		DW	XLOOP,	DTRA1-$
		DW	SEMIS

		DB	84H,'(."',')'+80h
	ALIGN 2
		DW	DTRAI - 12
PDOTQ		DW	DOCOL
		DW	RR
		DW	COUNT,	DUPE,	ONEP
		DW	FROMR,	PLUS,	TOR
		DW	TYPES,	SEMIS

		DB	0C2H,'.','"'+80h
	ALIGN 2
		DW	PDOTQ - 8
DOTQ		DW	DOCOL
		DW	LIT,	'"'
		DW	STATE,	JAT
		DW	ZBRAN,	DOTQ1-$
		DW	COMP
		DW	PDOTQ,	WORDS,	HERE
		DW	CAT,	ONEP,	ALLOT
		DW	BRAN,	DOTQ2-$
DOTQ1		DW	WORDS,	HERE,	COUNT,	TYPES
DOTQ2		DW	SEMIS

		DB	86H,"EXPEC","T"+80h
	ALIGN 2
		DW	DOTQ - 6
EXPEC		DW	DOCOL
		DW	OVER,	PLUS,	OVER
		DW	XDO
EXPE1		DW	KEY,	DUPE
		DW	LIT,	0EH
		DW	PORIG,	JAT,	EQUAL
		DW	ZBRAN,	EXPE2-$
		DW	DROP,	DUPE,	IDO
		DW	EQUAL,	DUPE,	FROMR
		DW	TWO,	SUBB,	PLUS
		DW	TOR
		DW	ZBRAN,	EXPE6-$
		DW	LIT,	BELL
		DW	BRAN,	EXPE7-$
EXPE6		DW	LIT,	BSOUT,	EMIT
		DW	BLS,	EMIT
		DW	LIT,	BSOUT
EXPE7		DW	BRAN,	EXPE3-$
EXPE2		DW	DUPE,	LIT,	ACR
		DW	EQUAL
		DW	ZBRAN,	EXPE4-$
		DW	LLEAVE,	DROP,	BLS,	ZERO
		DW	BRAN,	EXPE5-$
EXPE4		DW	DUPE
EXPE5		DW	IDO
		DW	CSTOR,	ZERO,	IDO,	ONEP
		DW	STORE
EXPE3		DW	EMIT
		DW	XLOOP,	EXPE1-$
		DW	DROP,	SEMIS

		DB	85H,"QUER","Y"+80h
		DW	EXPEC - 10
QUERY		DW	DOCOL
		DW	TIB,	JAT
		DW	LIT,	80,	EXPEC
		DW	ZERO,	INN,	STORE,	SEMIS

		DB	0C1H,80H
		DW	QUERY - 8
NULL		DW	DOCOL
		DW	BLK,	JAT
		DW	ZBRAN,	NULL1-$
		DW	ONE,	BLK,	PSTOR
		DW	ZERO,	INN,	STORE
		DW	BLK,	JAT
		DW	BSCR,	ONE,	SUBB,	ANDD
		DW	ZEQU
		DW	ZBRAN,	NULL2-$
		DW	QEXEC,	FROMR,	DROP
NULL2		DW	BRAN,	NULL3-$
NULL1		DW	FROMR,	DROP
NULL3		DW	SEMIS

		DB	84H,"FIL","L"+80h
	ALIGN 2
		DW	NULL - 4
FILL		DW	$ + 2
		POP	AX
		POP	CX
		POP	DI
		MOV	BX,DS
		MOV	ES,BX
		CLD
		REP STOSB
		JMP	NEXT

	ALIGN 2
		DB	85H,"ERAS","E"+80h
		DW	FILL - 8
ERASEE		DW	DOCOL
		DW	ZERO,	FILL,	SEMIS

		DB	86H,"BLANK","S"+80h
	ALIGN 2
		DW	ERASEE - 8
BLANK		DW	DOCOL
		DW	BLS,	FILL,	SEMIS

		DB	84H,"HOL","D"+80h
	ALIGN 2
		DW	BLANK - 10
HOLD		DW	DOCOL
		DW	LIT,	-1
		DW	HLD,	PSTOR
		DW	HLD,	JAT,	CSTOR,	SEMIS

		DB	83H,"PA","D"+80h
		DW	HOLD - 8
PAD		DW	DOCOL
		DW	HERE,	LIT,	68,	PLUS,	SEMIS
		DW	PLUS,	SEMIS

		DB	84H,"WOR","D"+80h
	ALIGN 2
		DW	PAD - 6
WORDS		DW	DOCOL
		DW	BLK,	JAT
		DW	ZBRAN,	WORD1-$
		DW	BLK,	JAT,	BLOCK
		DW	BRAN,	WORD2-$
WORD1		DW	TIB,	JAT
WORD2		DW	INN,	JAT,	PLUS,	SWAP
		DW	ENCL,	HERE
		DW	LIT,	34
		DW	BLANK,	INN,	PSTOR
		DW	OVER,	SUBB,	TOR
		DW	RR,	HERE,	CSTOR
		DW	PLUS,	HERE,	ONEP
		DW	FROMR,	CMOVE,	SEMIS

		DB	88H,"(NUMBER",")"+80h
	ALIGN 2
		DW	WORDS - 8
PNUMB		DW	DOCOL
PNUM1		DW	ONEP
		DW	DUPE,	TOR
		DW	CAT,	BASE,	JAT,	DIGIT
		DW	ZBRAN,	PNUM2-$
		DW	SWAP,	BASE,	JAT,	USTAR
		DW	DROP,	ROT,	BASE,	JAT
		DW	USTAR,	DPLUS
		DW	DPL,	JAT,	ONEP
		DW	ZBRAN,	PNUM3-$
		DW	ONE,	DPL,	PSTOR
PNUM3		DW	FROMR
		DW	BRAN,	PNUM1-$
PNUM2		DW	FROMR,	SEMIS

		DB	86H,"NUMBE","R"+80h
	ALIGN 2
		DW	PNUMB - 12
NUMB		DW	DOCOL
		DW	ZERO,	ZERO
		DW	ROT,	DUPE,	ONEP,	CAT
		DW	LIT,	"-",	EQUAL
		DW	DUPE,	TOR,	PLUS
		DW	LIT,	-1
NUMB1		DW	DPL,	STORE
		DW	PNUMB
		DW	DUPE,	CAT,	BLS,	SUBB
		DW	ZBRAN,	NUMB2-$
		DW	DUPE,	CAT
		DW	LIT,	".",	SUBB
		DW	ZERO,	QERR,	ZERO
		DW	BRAN,	NUMB1-$
NUMB2		DW	DROP,	FROMR
		DW	ZBRAN,	NUMB3-$
		DW	DMINU
	ALIGN 2
NUMB3		DW	SEMIS

		DB	85H,"-FIN","D"+80h
		DW	NUMB - 10
DFIND		DW	DOCOL
		DW	BLS,	WORDS
		DW	HERE,	CONT,	JAT,	JAT
		DW	PFIND,	DUPE,	ZEQU
		DW	ZBRAN,	DFIN1-$
		DW	DROP
		DW	HERE,	LATES,	PFIND
DFIN1		DW	SEMIS

		DB	87H,"(ABORT",")"+80h
		DW	DFIND - 8
PABOR		DW	DOCOL
		DW	ABORT,	SEMIS

		DB	85H,"ERRO","R"+80h
		DW	PABOR - 10
ERROR		DW	DOCOL
		DW	WARN,	JAT,	ZLESS
		DW	ZBRAN,	ERRO1-$
		DW	PABOR
ERRO1		DW	HERE,	COUNT,	TYPES
		DW	PDOTQ
		DB	2,"? "
		DW	MESS
		DW	SPSTO
		DW	BLK,	JAT,	DDUP
		DW	ZBRAN,	ERRO2-$
		DW	INN,	JAT,	SWAP
ERRO2		DW	QUIT

	ALIGN 2
		DB	83H,"ID","."+80h
		DW	ERROR - 8
IDDOT		DW	DOCOL
		DW	PAD
		DW	LIT,	32
		DW	LIT,	'_'
		DW	FILL
		DW	DUPE,	PFA,	LFA
		DW	OVER,	SUBB
		DW	PAD,	SWAP,	CMOVE
		DW	PAD,	COUNT
		DW	LIT,	1FH,	ANDD
; modified last character
		DW	TDUP,	PLUS,	ONEM
		DW	LIT,	-128,	SWAP,	PSTOR
; end of new lines
		DW	TYPES,	SPACE,	SEMIS

		DB	86H,"CREAT","E"+80h
	ALIGN 2
		DW	IDDOT - 6
CREAT		DW	DOCOL
		DW	DFIND
		DW	ZBRAN,	CREA1-$
		DW	DROP,	NFA,	IDDOT
		DW	LIT,	4,	MESS
		DW	SPACE
CREA1		DW	HERE,	DUPE,	CAT
		DW	WYDTH,	JAT,	MIN
		DW	ONEP,	ALLOT
		DW	DUPE
		DW	LIT,	0A0H
		DW	TOGGL
		DW	HERE,	ONE,	SUBB
		DW	LIT,	80H
		DW	TOGGL

		DW	DP,	JAT
		DW	ONEP
		DW	LIT,	0FFFEH,	ANDD
		DW	DP,	STORE

		DW	LATES,	COMMA
		DW	CURR,	JAT,	STORE
		DW	HERE,	TWOP,	COMMA,	SEMIS

		DB	0C9H,"[COMPILE","]"+80h
		DW	CREAT - 10
BCOMP		DW	DOCOL
		DW	DFIND
		DW	ZEQU,	ZERO,	QERR
		DW	DROP,	CFA,	COMMA,	SEMIS

		DB	0C7H,"LITERA","L"+80h
		DW	BCOMP - 12
LITER		DW	DOCOL
		DW	STATE,	JAT
		DW	ZBRAN,	LITE1-$
		DW	COMP,	LIT,	COMMA
LITE1		DW	SEMIS

		DB	0C8H,"DLITERA","L"+80h
	ALIGN 2
		DW	LITER - 10
DLITE		DW	DOCOL
		DW	STATE,	JAT
		DW	ZBRAN,	DLIT1-$
		DW	SWAP,	LITER,	LITER
DLIT1		DW	SEMIS

		DB	86H,"?STAC","K"+80h
	ALIGN 2
		DW	DLITE-12
QSTAC		DW	DOCOL
		DW	SPAT,	SZERO,	JAT
		DW	SWAP,	ULESS,	ONE,	QERR
		DW	SPAT,	HERE
		DW	LIT,	80H
		DW	PLUS,	ULESS
		DW	LIT,	7
		DW	QERR
		DW	SEMIS

		DB	89H,"INTERPRE","T"+80h
		DW	QSTAC - 10
INTER		DW	DOCOL
INTE1		DW	DFIND
		DW	ZBRAN,	INTE2-$
		DW	STATE,	 JAT,	LESS
		DW	ZBRAN,	INTE3-$
		DW	CFA,	COMMA
		DW	BRAN,	INTE4-$
INTE3		DW	CFA,	EXEC
INTE4		DW	QSTAC
		DW	BRAN,	INTE5-$
INTE2		DW	HERE,	NUMB,	DPL,	JAT,	ONEP
		DW	ZBRAN,	INTE6-$
		DW	DLITE
		DW	BRAN,	INTE7-$
INTE6		DW	DROP,	LITER
INTE7		DW	QSTAC
INTE5		DW	BRAN,	INTE1-$

		DB	89H,"IMMEDIAT","E"+80h
		DW	INTER-12
IMMED		DW	DOCOL
		DW	LATES
		DW	LIT,	40H
		DW	TOGGL,	SEMIS

		DB	8AH,"VOCABULAR","Y"+80h
	ALIGN 2
		DW	IMMED - 12
VOCAB		DW	DOCOL
		DW	BUILD
		DW	LIT,	0A081H
		DW	COMMA
		DW	CURR,	JAT
		DW	CFA,	COMMA,	HERE,	VOCL
		DW	JAT,	COMMA,	VOCL,	STORE
		DW	DOES
DOVOC		DW	TWOP,	CONT,	STORE,	SEMIS

		DB	0C5H,"FORT","H"+80h
		DW	VOCAB - 14
FORTH		DW	DODOE
		DW	DOVOC
		DW	0A081H
		DW	TASK - 8
		DW	0

		DB	8BH,"DEFINITION","S"+80h
		DW	FORTH - 8
DEFIN		DW	DOCOL
		DW	CONT,	JAT
		DW	CURR,	STORE,	SEMIS

		DB	0C1H,"("+80h
		DW	DEFIN - 14
PAREN		DW	DOCOL
		DW	LIT,	')',	WORDS,	SEMIS

		DB	84H,"QUI","T"+80h
	ALIGN 2
		DW	PAREN - 4
QUIT		DW	DOCOL
		DW	ZERO,	BLK,	STORE
		DW	LBRAC
QUIT1		DW	RPSTO,	CR,	QUERY
		DW	INTER
		DW	STATE,	JAT,	ZEQU
		DW	ZBRAN,	QUIT2-$
		DW	PDOTQ
		DB	2,"ok"
QUIT2		DW	BRAN,	QUIT1-$

	ALIGN 2
		DB	85H,"ABOR","T"+80h
		DW	QUIT - 8
ABORT		DW	DOCOL
		DW	SPSTO,	DECA,	QSTAC,	CR
		DW	DOTCPU,	PDOTQ
		DB	16H,'Fig-FORTH  Version '
		DB	FIGREL+30H, ADOT, FIGREV+30H
		DW	LIT,	10,	PORIG,	CAT
		DW	LIT,	41H,	PLUS,	EMIT
		DW	FORTH,	DEFIN
		DW	LIT,	0,	PRTER,	STORE	;Reset echo
		DW	QUIT

CTRLC:		
WRM:		MOV	SI,OFFSET WRM1
		JMP	NEXT
WRM1		DW	PABOR

	ALIGN 2
		DB	84H,"WAR","M"+80h
	ALIGN 2
		DW	ABORT - 8
WARM		DW	DOCOL
		DW	MTBUF,	ABORT

SYSINIT:	MOV	AX,CS
		MOV	DS,AX
		MOV	SS,AX
		MOV	ES,AX
		MOV	SP,WORD PTR [ORIG + 12H] ; set top of parameter stack
		MOV	SI,OFFSET COLDCFA ; "pointer" to COLD into "IP reg."
		CLD			; "direction upwards"
		MOV	BP,[RPP]	; set return stack empty (top=bottom)

		MOV	AH,25h	; DOS function "set interrupt vector"
		MOV	AL,23h	; interrupt number
		MOV	DX,OFFSET CTRLC ; addr of new interrupt handler
		INT	21h ; DOS Function Dispatcher (Set ^C exit address)
		JMP	NEXT

COLDCFA		DW	COLD

	ALIGN 2
		DB	84H,"COL","D"+80h
	ALIGN 2
		DW	WARM - 8
COLD		DW	DOCOL
		DW	MTBUF
		DW	ZERO,	DENSTY,	STORE
		DW	FIRST,	USE,	STORE
		DW	FIRST,	PREV,	STORE
		DW	DRZER
		DW	LIT,	ORIG+12H
		DW	LIT,	UP,	JAT
		DW	LIT,	6,	PLUS
		DW	LIT,	16,	CMOVE
		DW	LIT,	ORIG+12,JAT
		DW	LIT,	FORTH+6,STORE
		DW	LIT,	4,	SCR,	STORE
		DW	ABORT

		DB	84H,"S->","D"+80h
	ALIGN 2
		DW	COLD - 8
STOD		DW	$ + 2
		POP	DX
		SUB	AX,AX
		OR	DX,DX
		JNS	STOD1
		DEC	AX
STOD1:		JMP	DPUSH

	ALIGN 2
		DB	82H,"+","-"+80h
	ALIGN 2
		DW	STOD - 8
PM		DW	DOCOL
		DW	ZLESS
		DW	ZBRAN,	PM1-$
		DW	MINUS
PM1		DW	SEMIS

		DB	83H,"D+","-"+80h
		DW	PM - 6
DPM		DW	DOCOL
		DW	ZLESS
		DW	ZBRAN,	DPM1-$
		DW	DMINU
DPM1		DW	SEMIS

		DB	83H,"AB","S"+80h
		DW	DPM - 6
ABSO		DW	DOCOL
		DW	DUPE,	PM,	SEMIS

		DB	84H,"DAB","S"+80h
	ALIGN 2
		DW	ABSO - 6
DABS		DW	DOCOL
		DW	DUPE,	DPM,	SEMIS

		DB	83H,"MI","N"+80h
		DW	DABS - 8
MIN		DW	DOCOL
		DW	TDUP,	GREAT
		DW	ZBRAN,	MIN1-$
		DW	SWAP
MIN1		DW	DROP,	SEMIS

		DB	83H,"MA","X"+80h
		DW	MIN - 6
MAX		DW	DOCOL
		DW	TDUP,	LESS
		DW	ZBRAN,	MAX1-$
		DW	SWAP
MAX1		DW	DROP,	SEMIS

		DB	82H,"M","*"+80h
	ALIGN 2
		DW	MAX - 6
MSTAR		DW	DOCOL
		DW	TDUP,	XORR,	TOR
		DW	ABSO
		DW	SWAP,	ABSO,	USTAR
		DW	FROMR,	DPM,	SEMIS

		DB	82H,"M","/"+80h
	ALIGN 2
		DW	MSTAR - 6
MSLAS		DW	DOCOL
		DW	OVER,	TOR,	TOR
		DW	DABS
		DW	RR,	ABSO,	USLAS
		DW	FROMR,	RR,	XORR
		DW	PM,	SWAP,	FROMR
		DW	PM,	SWAP,	SEMIS

		DB	81H,"*"+80h
		DW	MSLAS - 6
STAR		DW	DOCOL
		DW	MSTAR,	DROP,	SEMIS

		DB	84H,"/MO","D"+80h
	ALIGN 2
		DW	STAR - 4
SLMOD		DW	DOCOL
		DW	TOR,	STOD,	FROMR
		DW	MSLAS,	SEMIS

		DB	81H,"/"+80h
		DW	SLMOD - 8
SLASH		DW	DOCOL
		DW	SLMOD,	SWAP,	DROP,	SEMIS

		DB	83H,"MO","D"+80h
		DW	SLASH - 4
MODD		DW	DOCOL
		DW	SLMOD,	DROP,	SEMIS

		DB	85H,"*/MO","D"+80h
		DW	MODD - 6
SSMOD		DW	DOCOL
		DW	TOR,	MSTAR,	FROMR
		DW	MSLAS,	SEMIS

		DB	82H,"*","/"+80h
	ALIGN 2
		DW	SSMOD - 8
SSLA		DW	DOCOL
		DW	SSMOD,	SWAP,	DROP,	SEMIS

		DB	85H,"M/MO","D"+80h
		DW	SSLA - 6
MSMOD		DW	DOCOL
		DW	TOR,	ZERO,	RR,	USLAS
		DW	FROMR,	SWAP,	TOR
		DW	USLAS,	FROMR,	SEMIS

		DB	86H,"(LINE",")"+80h
	ALIGN 2
		DW	MSMOD - 8
PLINE		DW	DOCOL
		DW	TOR
		DW	LIT,	64
		DW	BBUF,	SSMOD
		DW	FROMR,	BSCR,	STAR
		DW	PLUS
		DW	BLOCK,	PLUS
		DW	LIT,	64,	SEMIS

		DB	85H,".LIN","E"+80h
		DW	PLINE - 10
DLINE		DW	DOCOL
		DW	PLINE,	DTRAI,	TYPES,	SEMIS

		DB	87H,"MESSAG","E"+80h
		DW	DLINE - 8
MESS		DW	DOCOL
		DW	WARN,	JAT
		DW	ZBRAN,	MESS1-$
		DW	DDUP
		DW	ZBRAN,	MESS2-$
		DW	LIT,	4
		DW	OFSET,	JAT,	BSCR,	SLASH
		DW	SUBB,	DLINE,	SPACE
MESS2		DW	BRAN,	MESS3-$
MESS1		DW	PDOTQ
		DB	6,"MSG # "
		DW	DOT
MESS3		DW	SEMIS

	ALIGN 2
		DB	83H,"PC","@"+80h
		DW	MESS - 10
PTCAT		DW	$ + 2
		POP	DX
		IN	AL,DX		; was INB DX (?)
		SUB	AH,AH
		JMP	APUSH

	ALIGN 2
		DB	83H,"PC","!"+80h
		DW	PTCAT - 6
PTCSTO		DW	$ + 2
		POP	DX
		POP	AX
		OUT	DX,AL		; was OUTB DX (?)
		JMP	NEXT

	ALIGN 2
		DB	82H,"P","@"+80h
	ALIGN 2
		DW	PTCSTO - 6
PTAT		DW	$ + 2
		POP	DX
		IN	AX,DX		; was INW DX (?)
		JMP	APUSH

	ALIGN 2
		DB	82H,"P","!"+80h
	ALIGN 2
		DW	PTAT - 6
PTSTO		DW	$ + 2
		POP	DX
		POP	AX
		OUT	DX,AX		; was OUTW DX (?)
		JMP	NEXT

;		Disk Interface Words for MS-DOS, etc.
;		--------------------------------

	ALIGN 2
		DB	85H,"DRIV","E"+80h
		DW	PTSTO - 6
DRIVE		DW	DOVAR,	0

		DB	86H,"RECOR","D"+80h	;Not in fig listing
	ALIGN 2
		DW	DRIVE - 8
REC		DW	DOVAR,	0

		DB	83H,"US","E"+80h
		DW	REC - 10
USE		DW	DOVAR,	BUF1

		DB	84H,"PRE","V"+80h
	ALIGN 2
		DW	USE - 6
PREV		DW	DOVAR,	BUF1

		DB	87H,"SEC/BL","K"+80h
		DW	PREV - 8
SPBLK		DW	DOCON,	SPBL	; KBBUF / BPS

		DB	85H,"#BUF","F"+80h
		DW	SPBLK - 10
NOBUF		DW	DOCON,	NBUF

		DB	87H,"DENSIT","Y"+80h
		DW	NOBUF - 8
DENSTY		DW	DOVAR,	DODE

		DB	8AH,"DISK-ERRO","R"+80h
	ALIGN 2
		DW	DENSTY - 10
DSKERR		DW	DOVAR,	0

		DB	87H,"PRINTE","R"+80h		;EPRINT in fig
		DW	DSKERR - 14
PRTER		DW	DOVAR, 0

		DB	84H,"+BU","F"+80h
	ALIGN 2
		DW	PRTER - 10
PBUF		DW	DOCOL
		DW	BBUF,	TWOP,	TWOP	;B/BUF+4
		DW	PLUS,	DUPE,	LIMIT,	EQUAL
		DW	ZBRAN,	PBUF1-$
		DW	DROP,	FIRST
PBUF1		DW	DUPE,	PREV,	JAT
		DW	SUBB,	SEMIS

		DB	86H,"UPDAT","E"+80h
	ALIGN 2
		DW	PBUF - 8
UPDAT		DW	DOCOL
		DW	PREV,	JAT,	JAT
		DW	LIT,	8000H
		DW	ORR
		DW	PREV,	JAT,	STORE,	SEMIS

		DB	8DH,"EMPTY-BUFFER","S"+80h
		DW	UPDAT - 10
MTBUF		DW	DOCOL
		DW	FIRST,	LIMIT,	OVER
		DW	SUBB,	ERASEE,	SEMIS

		DB	83H,"DR","0"+80h
		DW	MTBUF - 16
DRZER		DW	DOCOL
		DW	ZERO,	OFSET,	STORE,	SEMIS

		DB	83H,"DR","1"+80h
		DW	DRZER - 6
DRONE		DW	DOCOL
		DW	DENSTY,	JAT
		DW	ZBRAN,	DRON1-$
		DW	LIT,	SPDRV2
		DW	BRAN,	DRON2-$
DRON1		DW	LIT,	SPDRV1
DRON2		DW	OFSET,	STORE,	SEMIS

		DB	86H,"BUFFE","R"+80h
	ALIGN 2
		DW	DRONE - 6
BUFFE		DW	DOCOL
		DW	USE,	JAT,	DUPE,	TOR
BUFF1		DW	PBUF
		DW	ZBRAN,	BUFF1-$
		DW	USE,	STORE
		DW	RR,	JAT,	ZLESS
		DW	ZBRAN,	BUFF2-$
		DW	RR,	TWOP
		DW	RR,	JAT
		DW	LIT,	7FFFH
		DW	ANDD,	ZERO,	RSLW
BUFF2		DW	RR,	STORE
		DW	RR,	PREV,	STORE
		DW	FROMR,	TWOP,	SEMIS

		DB	85H,"BLOC","K"+80h
		DW	BUFFE - 10
BLOCK		DW	DOCOL
		DW	OFSET,	JAT,	PLUS,	TOR
		DW	PREV,	JAT,	DUPE
		DW	JAT,	RR,	SUBB
		DW	DUPE,	PLUS
		DW	ZBRAN,	BLOC1-$
BLOC2		DW	PBUF,	ZEQU
		DW	ZBRAN,	BLOC3-$
		DW	DROP,	RR
		DW	BUFFE,	DUPE
		DW	RR,	ONE,	RSLW
		DW	TWO,	SUBB
BLOC3		DW	DUPE,	JAT,	RR,	SUBB
		DW	DUPE,	PLUS,	ZEQU
		DW	ZBRAN,	BLOC2-$
		DW	DUPE,	PREV,	STORE
BLOC1		DW	FROMR,	DROP
		DW	TWOP,	SEMIS

		DB	87H,"T&SCAL","C"+80h
		DW	BLOCK-8
TSCALC		DW	DOCOL
		DW	DENSTY,	JAT
		DW	ZBRAN,	TSCALS-$
		DW	LIT,	SPDRV2,	SLMOD
;		DW	LIT,	MXDRV,	MIN
		DW	DRIVE,	STORE
		DW	REC,	STORE,	SEMIS
;		single density calculations :
TSCALS		DW	LIT,	SPDRV1,	SLMOD
;		DW	LIT,	MXDRV,	MIN
		DW	DRIVE,	STORE
		DW	REC,	STORE,	SEMIS

		DB	8AH,"BLOCK-REA","D"+80h
	ALIGN 2
		DW	TSCALC - 10
BLKRD		DW	$ + 2
		MOV	[DSKERR+2],0	;reset error flag
		MOV	AX,[DRIVE+2]	;AL = drive no.
		MOV	BX,[USE+2]	;BX = transfer address
		MOV	CX,[SPBLK+2]	;CX = no. records to transfer
		MOV	DX,[REC+2]	;DX = logical record #
		PUSH	SI
		PUSH	BP
		INT	37		;BIOS disk read function
		JNC	READOK
;		MOV	B,[DSKERR+2],AL	;READ ERROR!  <- original line
		MOV	BYTE PTR [DSKERR+2],AL	;READ ERROR!
READOK:		POPF
		POP	BP
		POP	SI
		JMP	NEXT

	ALIGN 2
		DB	8BH,"BLOCK-WRIT","E"+80h
		DW	BLKRD - 14
BLKWRT		DW	$ + 2
		MOV	[DSKERR+2],0	;reset error flag
		MOV	AX,[DRIVE+2]
		MOV	BX,[USE+2]
		MOV	CX,[SPBLK+2]
		MOV	DX,[REC+2]
		PUSH	SI
		PUSH	BP
		INT	38		;BIOS disk write function
		JNC	WRTOK
		XOR	AH,AH		;return negative error code
		NEG	AX
		MOV	[DSKERR+2],AX	;WRITE ERROR!
WRTOK:		POPF
		POP	BP
		POP	SI
		JMP	NEXT

	ALIGN 2
		DB	83H,"R/","W"+80h
		DW	BLKWRT - 14
RSLW		DW	DOCOL
		DW	USE,	JAT,	TOR
		DW	TOR
		DW	SWAP,	USE, STORE
		DW	SPBLK,	STAR
		DW	TSCALC
		DW	FROMR
		DW	ZBRAN,	RSLW1-$
		DW	BLKRD
		DW	BRAN,	RSLW2-$
RSLW1		DW	BLKWRT
RSLW2		DW	FROMR,	USE,	STORE
		DW	DSKERR,	JAT,	DDUP
		DW	ZBRAN,	RSLW5-$		;OK
		DW	ZLESS
		DW	ZBRAN,	RSLW3-$
		DW	LIT,	9			;Write error
		DW	BRAN,	RSLW4-$
RSLW3		DW	LIT,	8			;Read error
RSLW4		DW	ZERO,	PREV,	JAT,	STORE	;This  buffer
							; is no good!
		DW	QERR
RSLW5		DW	SEMIS

		DB	85H,"FLUS","H"+80h
		DW	RSLW - 6
FLUSH		DW	DOCOL
		DW	NOBUF,	ONEP
		DW	ZERO,	XDO
FLUS1		DW	ZERO,	BUFFE,	DROP
		DW	XLOOP,	FLUS1-$
		DW	SEMIS

		DB	84H,"LOA","D"+80h
	ALIGN 2
		DW	FLUSH - 8
LOAD		DW	DOCOL
		DW	BLK,	JAT,	TOR
		DW	INN,	JAT,	TOR
		DW	ZERO,	INN,	STORE
		DW	BSCR,	STAR,	BLK,	STORE
		DW	INTER
		DW	FROMR,	INN,	STORE
		DW	FROMR,	BLK,	STORE
		DW	SEMIS

		DB	0C3H,"--",">"+80h
		DW	LOAD - 8
ARROW		DW	DOCOL
		DW	QLOAD
		DW	ZERO,	INN,	STORE
		DW	BSCR,	BLK,	JAT
		DW	OVER,	MODD,	SUBB
		DW	BLK,	PSTOR,	SEMIS

;****************************************
;*					*
;*	i/o primitives :		*
;*					*
;*	PQTER, PKEY, PEMIT, PCR,	*
;*	CONOUT, LSTOUT			*
;*					*
;****************************************

REQUEST		EQU	33		;BIOS function request intr.
CONOUT		EQU	2		;BIOS console output function
LSTOUT		EQU	5		;BIOS printer output function
CONIO		EQU	8		;BIOS console i/o fctn, no echo
CONSTAT		EQU	11		;BIOS console status check fctn

ACTRLC		EQU	3		;ASCII ^C

PQTER: 		MOV	AH,CONSTAT
		INT	REQUEST
		SUB	AH,AH
		JMP	APUSH

PKEY:		MOV	AX,1680h	; Program Idle (Release Timeslice)
		INT	2Fh		;     (let's be nice to CPU :)
		MOV	DX,0FFH
		MOV	AH,CONIO
		INT	REQUEST
		OR	AL,AL
		JZ	PKEY
;		AND	AX,7FH		; former "filter"
		AND	AX,0FFh		; pass all "high ASCII" characters (ZB)
		CMP	AL,ACTRLC	;check for ^C
		JNZ	PKEY1		;pass anything else
		INT	35		;Force ^C interrupt 
PKEY1:		JMP	APUSH

PEMIT		DW	$ + 2
		POP	DX
		CALL	POUT
		JMP	NEXT

PCR: 		MOV	DX,ACR
		CALL	POUT
		MOV	DX,LF
		CALL	POUT
		JMP	NEXT

; POUT: 		AND	DX,7FH
POUT:		MOV	AH,CONOUT
		INT	REQUEST
		MOV	BX,[ PRTER+2 ]	;Check echo flag
		OR	BX,BX
		JZ	OUTTA
		MOV	AH,LSTOUT
		INT	REQUEST		;Echo to printer
OUTTA:		RET

;********************************************************
;*							*
;*		TIME@, TIME!, DATE@, DATE!		*
;*							*
;********************************************************

	ALIGN 2
		DB	85H,"TIME","@"+80h
		DW	ARROW - 6
TIMAT		DW	$ + 2
		MOV	AH,2CH		;Get time
		INT	REQUEST
		PUSH	DX		;[sec sec/100]
		PUSH	CX		;[hr min]
		JMP	NEXT

	ALIGN 2
		DB	85H,"TIME","!"+80h
		DW	TIMAT - 8
TIMST		DW	$ + 2
		POP	CX		;[hr min]
		POP	DX		;[sec sec/100]
		MOV	AH,2DH
		INT	REQUEST
		JMP	NEXT

	ALIGN 2
		DB	85H,"DATE","@"+80h
		DW	TIMST - 8
DATAT		DW	$ + 2
		MOV	AH,2AH
		INT	REQUEST
		PUSH	CX		;year
		MOV	AL,DH		;month
		XOR	AH,AH
		XOR	DH,DH
		JMP	DPUSH		;DL=day

	ALIGN 2
		DB	85H,"DATE","!"+80h
		DW	DATAT - 8
DATST		DW	$ + 2
		POP	CX		;year
		POP	DX		;DL=day
		POP	AX
		MOV	DH,AL		;DH=month
		MOV	AH,2BH
		INT	REQUEST
		JMP	NEXT

EXIT: 		INT	32

	ALIGN 2
		DB	0C1H,"'"+80h
		DW	DATST - 8
TICK		DW	DOCOL
		DW	DFIND,	ZEQU
		DW	ZERO,	QERR
		DW	DROP,	LITER,	SEMIS

		DB	86H,"FORGE","T"+80h
	ALIGN 2
		DW	TICK - 4
FORG		DW	DOCOL
		DW	CURR,	JAT
		DW	CONT,	JAT
		DW	SUBB
		DW	LIT,	24,	QERR
		DW	TICK,	DUPE
		DW	FENCE,	JAT,	LESS
		DW	LIT,	21,	QERR
		DW	DUPE
		DW	NFA,	DP,	STORE
		DW	LFA,	JAT
		DW	CONT,	JAT,	STORE,	SEMIS

		DB	84H,"BAC","K"+80h
	ALIGN 2
		DW	FORG - 10
BACK		DW	DOCOL
		DW	HERE,	SUBB
		DW	COMMA,	SEMIS

		DB	0C5H,"BEGI","N"+80h
		DW	BACK - 8
BEGIN		DW	DOCOL
		DW	QCOMP
		DW	HERE,	ONE,	SEMIS

		DB	0C5H,"ENDI","F"+80h
		DW	BEGIN - 8
ENDIFF		DW	DOCOL
		DW	QCOMP
		DW	TWO,	QPAIR
		DW	HERE,	OVER,	SUBB
		DW	SWAP,	STORE,	SEMIS

		DB	0C4H,"THE","N"+80h
	ALIGN 2
		DW	ENDIFF - 8
THEN		DW	DOCOL
		DW	ENDIFF,	SEMIS

		DB	0C2H,"D","O"+80h
	ALIGN 2
		DW	THEN - 8
DO		DW	DOCOL
		DW	COMP,	XDO
		DW	HERE,	THREE,	SEMIS

		DB	0C4H,"LOO","P"+80h
	ALIGN 2
		DW	DO - 6
LOOPC		DW	DOCOL
		DW	THREE,	QPAIR
		DW	COMP,	XLOOP
		DW	BACK,	SEMIS

		DB	0C5H,"+LOO","P"+80h
		DW	LOOPC - 8
PLOOP		DW	DOCOL
		DW	THREE,	QPAIR
		DW	COMP,	XPLOO
		DW	BACK,	SEMIS

		DB	0C5H,"UNTI","L"+80h
		DW	PLOOP - 8
UNTIL		DW	DOCOL
		DW	ONE,	QPAIR
		DW	COMP,	ZBRAN
		DW	BACK,	SEMIS

		DB	0C3H,"EN","D"+80h
		DW	UNTIL - 8
ENDD		DW	DOCOL
		DW	UNTIL,	SEMIS

		DB	0C5H,"AGAI","N"+80h
		DW	ENDD - 6
AGAIN		DW	DOCOL
		DW	ONE,	QPAIR
		DW	COMP,	BRAN
		DW	BACK,	SEMIS

		DB	0C6H,"REPEA","T"+80h
	ALIGN 2
		DW	AGAIN - 8
REPEA		DW	DOCOL
		DW	TOR,	TOR
		DW	AGAIN
		DW	FROMR,	FROMR
		DW	TWO,	SUBB
		DW	ENDIFF,	SEMIS

		DB	0C2H,"I","F"+80h
	ALIGN 2
		DW	REPEA - 10
IFF		DW	DOCOL
		DW	COMP,	ZBRAN
		DW	HERE,	ZERO,	COMMA
		DW	TWO,	SEMIS

		DB	0C4H,"ELS","E"+80h
	ALIGN 2
		DW	IFF - 6
ELSEE		DW	DOCOL
		DW	TWO,	QPAIR
		DW	COMP,	BRAN
		DW	HERE,	ZERO,	COMMA
		DW	SWAP
		DW	TWO,	ENDIFF,	TWO
		DW	SEMIS

		DB	0C5H,"WHIL","E"+80h
		DW	ELSEE - 8
WHILE		DW	DOCOL
		DW	IFF,	TWOP,	SEMIS

		DB	86H,"SPACE","S"+80h
	ALIGN 2
		DW	WHILE - 8
SPACS		DW	DOCOL
		DW	ZERO,	MAX
		DW	DDUP
		DW	ZBRAN,	SPAX1-$
		DW	ZERO,	XDO
SPAX2		DW	SPACE
		DW	XLOOP,	SPAX2-$
SPAX1		DW	SEMIS

		DB	82H,"<","#"+80h
	ALIGN 2
		DW	SPACS - 10
BDIGS		DW	DOCOL
		DW	PAD,	HLD,	STORE
		DW	SEMIS

		DB	82H,"#",">"+80h
	ALIGN 2
		DW	BDIGS - 6
EDIGS		DW	DOCOL
		DW	DROP,	DROP
		DW	HLD,	JAT
		DW	PAD
		DW	OVER,	SUBB,	SEMIS

		DB	84H,"SIG","N"+80h
	ALIGN 2
		DW	EDIGS - 6
SIGN		DW	DOCOL
		DW	ROT,	ZLESS
		DW	ZBRAN,	SIGN1-$
		DW	LIT,	'-',	HOLD
SIGN1		DW	SEMIS

		DB	81H,"#"+80h
		DW	SIGN - 8
DIG		DW	DOCOL
		DW	BASE,	JAT,	MSMOD
		DW	ROT
		DW	LIT,	9
		DW	OVER,	LESS
		DW	ZBRAN,	DIG1-$
		DW	LIT,	7,	PLUS
DIG1		DW	LIT,	'0',	PLUS
		DW	HOLD,	SEMIS

		DB	82H,"#","S"+80h
	ALIGN 2
		DW	DIG - 4
DIGS		DW	DOCOL
DIGS1		DW	DIG
		DW	OVER,	OVER
		DW	ORR,	ZEQU
		DW	ZBRAN,	DIGS1-$
		DW	SEMIS

		DB	83H,"D.","R"+80h
		DW	DIGS - 6
DDOTR		DW	DOCOL
		DW	TOR,	SWAP,	OVER
		DW	DABS
		DW	BDIGS
		DW	DIGS,	SIGN
		DW	EDIGS
		DW	FROMR,	OVER,	SUBB
		DW	SPACS,	TYPES,	SEMIS

		DB	82H,".","R"+80h
	ALIGN 2
		DW	DDOTR - 6
DOTR		DW	DOCOL
		DW	TOR
		DW	STOD,	FROMR,	DDOTR,	SEMIS

		DB	82H,"D","."+80h
	ALIGN 2
		DW	DOTR - 6
DDOT		DW	DOCOL
		DW	ZERO
		DW	DDOTR,	SPACE,	SEMIS

		DB	81H,"."+80h
		DW	DDOT - 6
DOT		DW	DOCOL
		DW	STOD,	DDOT,	SEMIS

		DB	81H,"?"+80h
		DW	DOT - 4
QUES		DW	DOCOL
		DW	JAT,	DOT,	SEMIS

		DB	82H,"U","."+80h
	ALIGN 2
		DW	QUES - 4
UDOT		DW	DOCOL
		DW	ZERO,	DDOT,	SEMIS

		DB	85H,"VLIS","T"+80h
		DW	UDOT - 6
VLIST		DW	DOCOL
		DW	LIT,	80H
		DW	OUTT,	STORE
		DW	CONT,	JAT,	JAT
VLIS1		DW	OUTT,	JAT
		DW	CSLL,	GREAT
		DW	ZBRAN,	VLIS2-$
		DW	CR
		DW	ZERO,	OUTT,	STORE
VLIS2		DW	DUPE
		DW	IDDOT
		DW	SPACE,	SPACE
		DW	PFA,	LFA,	JAT
		DW	DUPE,	ZEQU
		DW	QTERM,	ORR
		DW	ZBRAN,	VLIS1-$
		DW	DROP,	SEMIS

		DB	83H,"BY","E"+80h
		DW	VLIST - 8
BYE		DW	$ + 2
		JMP	EXIT

	ALIGN 2
		DB	84H,"LIS","T"+80h
	ALIGN 2
		DW	BYE - 6
LISTC		DW	DOCOL
		DW	DECA,	CR
		DW	DUPE,	SCR,	STORE
		DW	PDOTQ
		DB	6,"SCR # "
		DW	DOT
		DW	LIT,	16,	ZERO,	XDO
LIST1		DW	CR,	IDO
		DW	LIT,	3,	DOTR,	SPACE
		DW	IDO,	SCR,	JAT,	DLINE
		DW	QTERM
		DW	ZBRAN,	LIST2-$
		DW	LLEAVE
LIST2		DW	XLOOP,	LIST1-$
		DW	CR,	SEMIS

	ALIGN 2
		DB	85H,"INDE","X"+80h
		DW	LISTC - 8
INDEX		DW	DOCOL
		DW	LIT,	FF,	EMIT,	CR
		DW	ONEP,	SWAP,	XDO
INDE1		DW	CR,	IDO
		DW	LIT,	3,	DOTR,	SPACE
		DW	ZERO,	IDO,	DLINE
		DW	QTERM
		DW	ZBRAN,	INDE2-$
		DW	LLEAVE
INDE2		DW	XLOOP,	INDE1-$
		DW	SEMIS

		DB	85H,"TRIA","D"+80h
		DW	INDEX - 8
TRIAD		DW	DOCOL
		DW	LIT,	FF,	EMIT
		DW	LIT,	3,	SLASH
		DW	LIT,	3,	STAR
		DW	LIT,	3,	OVER
		DW	PLUS,	SWAP,	XDO
TRIA1		DW	CR,	IDO,	LISTC
		DW	QTERM
		DW	ZBRAN,	TRIA2-$
		DW	LLEAVE
TRIA2		DW	XLOOP,	TRIA1-$
		DW	CR
		DW	LIT,	15,	MESS,	CR
		DW	SEMIS

		DB	84H,".CP","U"+80h
	ALIGN 2
		DW	TRIAD - 8
DOTCPU		DW	DOCOL
		DW	BASE,	JAT
		DW	LIT,	36,	BASE,	STORE
		DW	LIT,	22H,	PORIG,	TAT
		DW	DDOT
		DW	BASE,	STORE,	SEMIS

		DB	85H,"MATC","H"+80h
		DW	DOTCPU - 8
MATCH		DW	$ + 2
		MOV	DI,SI
		POP	CX
		POP	BX
		POP	DX
		POP	SI
		PUSH	SI
MAT1: 		LODSB
		CMP	AL,[BX]
		JNZ	MAT3
		PUSH	BX
		PUSH	CX
		PUSH	SI
MAT2: 		DEC	CX
		JZ	MATCHOK
		DEC	DX
		JZ	NOMATCH
		INC	BX
		LODSB
		CMP	AL,[BX]
		JZ	MAT2
		POP	SI
		POP	CX
		POP	BX
MAT3: 		DEC	DX
		JNZ	MAT1
		JMP	MAT4
MATCHOK:
NOMATCH: 	POP	CX
		POP	CX
		POP	CX
MAT4: 		MOV	AX,SI
		POP	SI
		SUB	AX,SI
		MOV	SI,DI
		JMP	DPUSH

	ALIGN 2
		DB	82H,".","S"+80h
	ALIGN 2
		DW	MATCH - 8
DOTES		DW	DOCOL
		DW	LIT,	2
		DW	SPACS,	SPAT,	SZERO
		DW	JAT,	EQUAL,	ZEQU
		DW	ZBRAN,	DOTES1 - $
		DW	SPAT,	SZERO,	JAT
		DW	TWOM,	XDO
DOTES2		DW	IDO,	JAT,	DOT
		DW	LIT,	-2,	XPLOO,	DOTES2 - $
		DW	LIT,	2,	SPACS
DOTES1		DW	SEMIS

		DB	84H,"PAG","E"+80h
	ALIGN 2
		DW	DOTES - 6
PAGEE		DW	$ + 2
		MOV	AX,03h
		INT	10h
		JMP	NEXT
	ALIGN 2
		DB	85H,"AT-X","Y"+80h
		DW	PAGEE - 8
ATXY 		DW	$ + 2
		MOV	AX,INITS0	; get bottom of the stack address
		MOV	BX,SP		; get current stack level
		SUB	AX,4		; decrement that bottom addr value...
		CMP	AX,BX		; ...and check is it still >= SP
		JNGE	ATXYERROR	; lesser means "not enough parameters"
		POP	AX
		MOV	DH,AL
		POP	AX
		MOV	DL,AL
		XOR	AX,AX
		MOV	AH,2
		MOV	BH,0
		INT	10h
		JMP	NEXT
ATXYERROR:	POP	AX		; generate error
		POP	AX
		JMP	NEXT
	ALIGN 2
		DB	86H,"GET-X","Y"+80h
	ALIGN 2
		DW	ATXY - 8
GETXY		DW	$ + 2
		MOV	AH,0Fh	; get current video mode
		INT	10h	; BIOS video services
		MOV	BL,0	; (perhaps not necessary?)
		MOV	AH,03	; get cursor position (page# already in BH)
		INT	10h
		XOR	AX,AX
		MOV	AL,DL
		PUSH	AX
		MOV	AL,DH
		PUSH	AX
		JMP	NEXT

	ALIGN 2
		DB	84H,"MSE","C"+80h
	ALIGN 2
		DW	GETXY - 10
DELAY		DW	$ + 2
		MOV	AX,INITS0	; get bottom of the stack address
		MOV	BX,SP		; get current stack level
		CMP	AX,BX		; ...and check is it still >= SP
		JZ	DELAYERROR	; equal means "not enough parameters"
		MOV	AX,1680h	; Program Idle (Release Timeslice)
		INT	2Fh		;     (let's be nice to CPU :)
		MOV	AH,86h		; BIOS "wait" function
		XOR	CX,CX
		POP	DX
		MOV	CL,DH
		MOV	DH,DL
		XOR	DL,DL
		SHL	CX,1
		SHL	DX,1
		ADC	CX,0
		SHL	CX,1
		SHL	DX,1
		ADC	CX,0
		INT	15h
		XOR	AX,AX
		ADC	AX,0
		JMP	APUSH
DELAYERROR:	POP	AX		; generate error
		JMP	NEXT


	ALIGN 2
		DB	84H,"TAS","K"+80h
	ALIGN 2
		DW	DELAY - 8
TASK		DW	DOCOL
		DW	SEMIS

INITDP		EQU	$
main		ENDP

codeSegm	ENDS
; ---------------------------
		END	main

